from typing import Dict, Optional

from melange.examples.saga_pattern.models import Saga


class SagaRepository:
    def __init__(self) -> None:
        self.sagas: Dict[str, Saga] = {}

    def save(self, payment: Saga) -> None:
        self.sagas[payment.id] = payment

    def get(self, id: str) -> Optional[Saga]:
        return self.sagas.get(id)



import os

from melange.backends.sqs.localsqs import LocalSQSBackend
from melange.examples.payment_service.consumer import PaymentConsumer
from melange.examples.payment_service.publisher import PaymentPublisher
from melange.examples.payment_service.repository import PaymentRepository
from melange.examples.payment_service.service import PaymentService
from melange.examples.shared import serializer_registry
from melange.message_dispatcher import SimpleMessageDispatcher
from melange.publishers import QueuePublisher
from melange.serializers.pickle import PickleSerializer

if __name__ == "__main__":
    serializer = PickleSerializer()
    backend = LocalSQSBackend(
        host=os.environ.get("SQSHOST"), port=os.environ.get("SQSPORT")
    )

    payment_consumer = SimpleMessageDispatcher(
        PaymentConsumer(
            PaymentService(
                PaymentRepository(),
                PaymentPublisher(QueuePublisher(serializer_registry, backend=backend)),
            )
        ),
        serializer_registry,
        backend=backend,
    )

    print("Consuming...")
    payment_consumer.consume_loop("payment-updates")



from melange.backends.factory import MessagingBackendFactory
from melange.backends.sqs.sqs_backend import AWSBackend

backend = AWSBackend()
factory = MessagingBackendFactory(backend)
factory.init_topic("my-topic")



# type: ignore
"""
Looking for maintainers for the RabbitMQBackend!
"""


import json

import pika

from melange.backends.interfaces import MessagingBackend
from melange.models import Message


class RabbitMQBackend(MessagingBackend):
    def __init__(self, **kwargs):
        super().__init__()
        connection_parameters = pika.ConnectionParameters(**kwargs)
        self.connection = pika.BlockingConnection(connection_parameters)
        self.channel = self.connection.channel()
        self.channel.basic_qos(prefetch_count=1)

    def retrieve_messages(self, queue):
        method_frame, properties, body = self.channel.basic_get(
            queue=queue, no_ack=False
        )

        if not method_frame:
            return []

        message = Message(
            message_id=str(method_frame.delivery_tag),
            content=json.loads(body.decode("utf-8")),
            metadata=properties,
        )
        return [message]

    def publish_to_topic(self, content, topic, event_type_name):
        result = self.channel.basic_publish(
            exchange=topic,
            routing_key="",
            body=content,
            properties=pika.BasicProperties(
                delivery_mode=2,  # make message persistent
            ),
        )

        return result

    def get_queue(self, queue_name):
        queue = self.channel.queue_declare(queue=queue_name, durable=True)
        return queue.method.queue

    def declare_topic(self, topic_name):
        self.channel.exchange_declare(
            exchange=topic_name, exchange_type="fanout", durable=True
        )

        return topic_name

    def acknowledge(self, message):
        self.channel.basic_ack(delivery_tag=int(message.message_id))

    def declare_queue(self, queue_name, *topics_to_bind, dead_letter_queue_name=None):
        queue = self.channel.queue_declare(queue=queue_name, durable=True)
        if topics_to_bind:
            for topic in topics_to_bind:
                self.channel.queue_bind(exchange=topic, queue=queue_name)

        return queue.method.queue, dead_letter_queue_name

    def close_connection(self):
        self.connection.close()

    def delete_queue(self, queue):
        self.channel.queue_delete(queue=queue)

    def delete_topic(self, topic):
        self.channel.exchange_delete(exchange=topic)



from simple_cqrs.domain_event import DomainEvent

from melange.backends.sqs.localsqs import LocalSQSBackend
from melange.consumers import SingleDispatchConsumer, consumer
from melange.examples.doc_examples.tutorial.publish import MyTestMessage
from melange.message_dispatcher import SimpleMessageDispatcher
from melange.serializers.pickle import PickleSerializer
from melange.serializers.registry import SerializerRegistry


class MyConsumer(SingleDispatchConsumer):
    @consumer
    def on_my_test_message_received(self, event: MyTestMessage) -> None:
        print(event.message)


if __name__ == "__main__":
    serializer_settings = {
        "serializers": {"pickle": PickleSerializer},
        "serializer_bindings": {DomainEvent: "pickle"},
    }

    serializer_registry = SerializerRegistry(serializer_settings)

    backend = LocalSQSBackend(host="localhost", port=9324)
    serializer = PickleSerializer()
    consumer = MyConsumer()
    message_dispatcher = SimpleMessageDispatcher(
        consumer,
        serializer_registry,
        backend=backend,
    )
    print("Consuming...")
    message_dispatcher.consume_loop("melangetutorial-queue")



import codecs
import pickle
from typing import Optional

from simple_cqrs.domain_event import DomainEvent

from melange.serializers.interfaces import Serializer


class PickleSerializer(Serializer[DomainEvent]):
    """
    Serializes DomainEvents with pickle.
    """

    @staticmethod
    def identifier() -> int:
        return 2

    def manifest(self, data: DomainEvent) -> str:
        return data.__class__.__qualname__

    def deserialize(self, data: str, manifest: Optional[str] = None) -> DomainEvent:
        return pickle.loads(codecs.decode(data.encode(), "base64"))

    def serialize(self, data: DomainEvent) -> str:
        return codecs.encode(pickle.dumps(data), "base64").decode()



from typing import Any, List, Optional

from melange.backends.backend_manager import BackendManager
from melange.backends.interfaces import MessagingBackend
from melange.models import QueueWrapper, TopicWrapper


class MessagingBackendFactory:
    """
    Class used to create queues and topics
    """

    def __init__(self, backend: Optional[MessagingBackend] = None) -> None:
        self._backend = backend or BackendManager().get_default_backend()

    def init_queue(
        self,
        queue_name: str,
        topic_names_to_subscribe: Optional[List[str]] = None,
        dlq_name: Optional[str] = None,
        **kwargs: Any
    ) -> QueueWrapper:
        """
        Creates a queue if it does not exist, subscribes it to the topics, and creates a
        dead letter queue.

        Args:
            queue_name: name of the queue to create
            topic_names_to_subscribe: list of topic names to subscribe this queue. If
                the topics do not exist they will be created.
            dlq_name: name of the dead letter queue to create and attach to the main queue
            **kwargs: any other arguments required by your messaging backend

        Returns:
            The created queue wrapper

        """
        topic_names_to_subscribe = topic_names_to_subscribe or []
        topics = [self.init_topic(t) for t in topic_names_to_subscribe]

        queue, dlq = self._backend.declare_queue(
            queue_name, *topics, dead_letter_queue_name=dlq_name, **kwargs
        )

        return queue

    def init_topic(self, topic_name: str) -> TopicWrapper:
        """
        Creates a topic if it does not exist.
        Args:
            topic_name: the name of the topic to create

        Returns:
            The created topic wrapper

        """
        return self._backend.declare_topic(topic_name)



import logging
from functools import reduce
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type

import funcy
from toolz import dicttoolz, itertoolz

from melange.serializers.interfaces import Serializer
from melange.serializers.json import JsonSerializer
from melange.serializers.pickle import PickleSerializer

logger = logging.getLogger(__name__)

default_settings = {
    "serializers": {"json": JsonSerializer, "pickle": PickleSerializer},
    "serializer_bindings": {Dict: "json"},
    "default": "pickle",
}

ClassSerializer = Tuple[Type, Type[Serializer]]


class SerializerRegistry:
    """
    This serializer registry tries to match the interface of akka:
    https://github.com/akka/akka/blob/main/akka-actor/src/main/scala/akka/serialization/Serialization.scala

    Explanations: https://doc.akka.io/docs/akka/current/serialization.html
    """

    def __init__(self, config: Dict):
        self._config_serializers: Dict[str, Type[Serializer]] = config["serializers"]
        self._config_bindings: Dict[Type, str] = config["serializer_bindings"]

        self._serializer_map: Dict[Type, Type[Serializer]] = dicttoolz.valmap(
            lambda v: self._config_serializers[v], self._config_bindings
        )

        # The bindings are sorted by specifity (meaning the lower the types
        # in the hierarchy occupy the first positions)
        self._bindings = list(sort((k, v) for k, v in self._serializer_map.items()))

        self._default_serializer: Optional[Type[Serializer]]
        try:
            self._default_serializer = self._config_serializers[config["default"]]
        except Exception:
            self._default_serializer = None

        self._serializers_by_id: Dict[int, Type[Serializer]] = {
            v.identifier(): v for k, v in self._config_serializers.items()  # type: ignore
        }
        self._quickserializer_by_identity: List[Optional[Type[Serializer]]] = [
            None
        ] * 1024

        for k, v in self._serializers_by_id.items():
            self._quickserializer_by_identity[k] = v

    def get_serializer_by_id(self, id: int) -> Type[Serializer]:
        if id < 1024:
            serializer = self._quickserializer_by_identity[id]
            if not serializer:
                raise Exception(f"Serializer with ID {id} not found")
            return serializer
        else:
            return self._serializers_by_id[id]

    def get_serializer_by_name(self, name: str) -> Type[Serializer]:
        return self._config_serializers[name]

    def deserialize_with_serializerid(
        self, data: str, serializer_id: int, manifest: Optional[str]
    ) -> Any:
        serializer = self.get_serializer_by_id(serializer_id)()
        return serializer.deserialize(data, manifest)

    def deserialize_with_class(
        self, data: str, klass: Type, manifest: Optional[str]
    ) -> Any:
        serializer = self.serializer_for(klass)()
        return serializer.deserialize(data, manifest)

    def serialize(self, obj: Any) -> str:
        serializer = self.find_serializer_for(obj)
        data = serializer.serialize(obj)
        return data

    def find_serializer_for(self, obj: Any) -> Serializer:
        return self.serializer_for(type(obj))()

    def serializer_for(self, type: Type) -> Type[Serializer]:
        if self._serializer_map.get(type):
            return self._serializer_map[type]
        else:
            # Check if the type is a subclass of any of the defined serializers
            possible_bindings = funcy.lfilter(
                lambda t: issubclass(type, t), self._bindings
            )
            if len(possible_bindings) == 0:
                # No serializer found. Return the default serializer
                if self._default_serializer:
                    return self._default_serializer
                else:
                    raise Exception(f"No serializer could be found for the type {type}")
            elif len(possible_bindings) == 1:
                return possible_bindings[0][1]
            else:
                logger.warning(
                    f"More than one serializer found for type {type}. "
                    "Choosing the first one."
                )
                return itertoolz.first(possible_bindings)[1]


def sort(list_items: Iterable[ClassSerializer]) -> Iterable[ClassSerializer]:
    """
    Sort so that subtypes always precede their supertypes, but without
    obeying any order between unrelated subtypes (insert sort).
    """

    def _(
        buffer: List[ClassSerializer], type_tuple: ClassSerializer
    ) -> List[ClassSerializer]:
        t = type_tuple[0]

        index = next((i for i, v in enumerate(buffer) if issubclass(t, v[0])), None)
        if index is None:
            buffer.append(type_tuple)
        else:
            buffer.insert(index, type_tuple)

        return buffer

    init: List[ClassSerializer] = []
    return reduce(_, list(list_items), init)  # type: ignore



import weakref
from typing import Any, Dict, Iterable, List, Optional, Tuple

from melange.models import Message, QueueWrapper, TopicWrapper


class MessagingBackend:
    def __init__(self) -> None:
        self._finalizer = weakref.finalize(self, self.close_connection)

    def declare_topic(self, topic_name: str) -> TopicWrapper:
        """
        Gets or creates a topic.

        Args:
            topic_name: The name of the topic to create

        Returns:
            An object that represents a topic. The type of the object
            is only relevant inside the context of the backend, so what you
            return as a topic will be passed in next calls to the backend
            where a topic is required
        """
        raise NotImplementedError

    def get_queue(self, queue_name: str) -> QueueWrapper:
        """
        Gets the queue with the name `queue_name`. Does not perform creation.

        Args:
            queue_name: the name of the queue to retrieve

        Returns:
            A `Queue` object that represents the created the queue
        """
        raise NotImplementedError

    def declare_queue(
        self,
        queue_name: str,
        *topics_to_bind: TopicWrapper,
        dead_letter_queue_name: Optional[str] = None,
        **kwargs: Any
    ) -> Tuple[QueueWrapper, Optional[QueueWrapper]]:
        """
        Gets or creates a queue.

        Args:
            queue_name: the name of the queue to create
            *topics_to_bind: if provided, creates all these topics and subscribes
                the created queue to them
            dead_letter_queue_name: if provided, create a dead letter queue attached to
                the created `queue_name`.
            **kwargs:

        Returns:
            A tuple with the created queue and the dead letter queue (if applies)
        """

        raise NotImplementedError

    def retrieve_messages(self, queue: QueueWrapper, **kwargs: Any) -> List[Message]:
        """
        Retrieves a list of available messages from the queue.

        Args:
            queue: the queue object
            **kwargs: Other parameters/options required by the backend

        Returns:
            A list of available messages from the queue
        """
        raise NotImplementedError

    def yield_messages(self, queue: QueueWrapper, **kwargs: Any) -> Iterable[Message]:
        """
        Yields available messages from the queue.

        Args:
            queue: the queue object
            **kwargs: Other parameters/options required by the backend

        Returns:
            An iterable which will poll the queue upon requesting more messages
        """
        raise NotImplementedError

    def publish_to_topic(
        self,
        message: Message,
        topic: TopicWrapper,
        extra_attributes: Optional[Dict] = None,
    ) -> None:
        """
        Publishes a message content and the manifest to the topic

        Args:
            message: the message to send
            topic: the topic to send the message to
            extra_attributes: extra properties that might be required for the backend
        """
        raise NotImplementedError

    def publish_to_queue(
        self, message: Message, queue: QueueWrapper, **kwargs: Any
    ) -> None:
        raise NotImplementedError

    def acknowledge(self, message: Message) -> None:
        """
        Acknowledges a message so that it won't be redelivered by
        the messaging infrastructure in the future
        """
        raise NotImplementedError

    def close_connection(self) -> None:
        """
        Override this function if you want to use some finalizer code
         to shutdown your backend in a clean way
        """
        pass

    def delete_queue(self, queue: QueueWrapper) -> None:
        """
        Deletes the queue
        """
        raise NotImplementedError

    def delete_topic(self, topic: TopicWrapper) -> None:
        """
        Deletes the topic
        """
        raise NotImplementedError



from datetime import datetime
from typing import Any, Optional

from simple_cqrs.domain_event import DomainEvent


class OrderResponse(DomainEvent):
    def __init__(
        self, id: str, reference: str, occurred_on: Optional[datetime] = None
    ) -> None:
        super().__init__(occurred_on)
        self.id = id
        self.reference = reference


class OrderStatus(DomainEvent):
    def __init__(
        self,
        order_id: str,
        payment_id: str,
        status: str,
        occurred_on: Optional[datetime] = None,
        **kwargs: Any
    ) -> None:
        super().__init__(occurred_on, **kwargs)
        self.order_id = order_id
        self.payment_id = payment_id
        self.status = status



import json
import logging
import uuid
from json import JSONDecodeError
from typing import Any, Dict, Iterable, List, Optional, Tuple

import boto3

from melange.backends.interfaces import MessagingBackend
from melange.models import Message, QueueWrapper, TopicWrapper

logger = logging.getLogger(__name__)


class BaseSQSBackend(MessagingBackend):
    """
    Base class for SQS Backends
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self.max_number_of_messages = kwargs.get("max_number_of_messages", 10)
        self.visibility_timeout = kwargs.get("visibility_timeout", 100)
        self.wait_time_seconds = kwargs.get("wait_time_seconds", 10)

        self.extra_settings = kwargs.get("extra_settings", {})
        self.sns_settings = kwargs.get("sns_settings", {})

    def declare_topic(self, topic_name: str) -> TopicWrapper:
        sns = boto3.resource("sns", **self.sns_settings)
        topic = sns.create_topic(Name=topic_name)
        return TopicWrapper(topic)

    def get_queue(self, queue_name: str) -> QueueWrapper:
        sqs_res = boto3.resource("sqs", **self.extra_settings)
        return QueueWrapper(sqs_res.get_queue_by_name(QueueName=queue_name))

    def _subscribe_to_topics(
        self, queue: QueueWrapper, topics_to_bind: Iterable[TopicWrapper], **kwargs: Any
    ) -> None:
        if topics_to_bind:
            statements = []
            for topic in topics_to_bind:
                statement = {
                    "Sid": "Sid{}".format(uuid.uuid4()),
                    "Effect": "Allow",
                    "Principal": "*",
                    "Resource": queue.unwrapped_obj.attributes["QueueArn"],
                    "Action": "sqs:SendMessage",
                    "Condition": {
                        "ArnEquals": {"aws:SourceArn": topic.unwrapped_obj.arn}
                    },
                }

                statements.append(statement)

                subscription = topic.unwrapped_obj.subscribe(
                    Protocol="sqs",
                    Endpoint=queue.unwrapped_obj.attributes[
                        "QueueArn"
                    ],  # , Attributes={"RawMessageDelivery": "true"}
                )

                if kwargs.get("filter_events"):
                    filter_policy = {"manifest": kwargs["filter_events"]}
                else:
                    filter_policy = {}

                subscription.set_attributes(
                    AttributeName="FilterPolicy",
                    AttributeValue=json.dumps(filter_policy),
                )

            policy = {
                "Version": "2012-10-17",
                "Id": "sqspolicy",
                "Statement": statements,
            }

            queue.unwrapped_obj.set_attributes(
                Attributes={"Policy": json.dumps(policy)}
            )

    def declare_queue(
        self,
        queue_name: str,
        *topics_to_bind: TopicWrapper,
        dead_letter_queue_name: Optional[str] = None,
        **kwargs: Any,
    ) -> Tuple[QueueWrapper, Optional[QueueWrapper]]:
        try:
            queue = self.get_queue(queue_name)
        except Exception:
            queue = self._create_queue(queue_name, content_based_deduplication="true")

        self._subscribe_to_topics(queue, topics_to_bind, **kwargs)

        dead_letter_queue: Optional[QueueWrapper] = None
        if dead_letter_queue_name:
            try:
                dead_letter_queue = self.get_queue(dead_letter_queue_name)
            except Exception:
                dead_letter_queue = self._create_queue(
                    dead_letter_queue_name, content_based_deduplication="true"
                )

            redrive_policy = {
                "deadLetterTargetArn": dead_letter_queue.unwrapped_obj.attributes[
                    "QueueArn"
                ],
                "maxReceiveCount": "4",
            }

            queue.unwrapped_obj.set_attributes(
                Attributes={"RedrivePolicy": json.dumps(redrive_policy)}
            )

        return queue, dead_letter_queue

    def _create_queue(self, queue_name: str, **kwargs: Any) -> QueueWrapper:
        sqs_res = boto3.resource("sqs", **self.extra_settings)
        fifo = queue_name.endswith(".fifo")
        attributes = {}
        if fifo:
            attributes["FifoQueue"] = "true"
            attributes["ContentBasedDeduplication"] = (
                "true" if kwargs.get("content_based_deduplication") else "false"
            )
        queue = sqs_res.create_queue(QueueName=queue_name, Attributes=attributes)
        return QueueWrapper(queue)

    def retrieve_messages(self, queue: QueueWrapper, **kwargs: Any) -> List[Message]:
        args = dict(
            MaxNumberOfMessages=self.max_number_of_messages,
            VisibilityTimeout=self.visibility_timeout,
            WaitTimeSeconds=self.wait_time_seconds,
            MessageAttributeNames=["All"],
            AttributeNames=["All"],
        )

        if "attempt_id" in kwargs:
            args["ReceiveRequestAttemptId"] = kwargs["attempt_id"]

        messages = queue.unwrapped_obj.receive_messages(**args)

        # We need to differentiate here whether the message came from SNS or SQS
        return [self._construct_message(message) for message in messages]

    def yield_messages(self, queue: QueueWrapper, **kwargs: Any) -> Iterable[Message]:
        args = dict(
            MaxNumberOfMessages=self.max_number_of_messages,
            VisibilityTimeout=self.visibility_timeout,
            WaitTimeSeconds=self.wait_time_seconds,
            MessageAttributeNames=["All"],
            AttributeNames=["All"],
        )

        while True:
            messages = queue.unwrapped_obj.receive_messages(**args)
            for message_content in messages:
                message = self._construct_message(message_content)
                yield message

    def publish_to_queue(
        self, message: Message, queue: QueueWrapper, **kwargs: Any
    ) -> None:
        message_args: Dict = {}
        message_args["MessageBody"] = json.dumps({"Message": message.content})

        is_fifo = queue.unwrapped_obj.attributes.get("FifoQueue") == "true"
        message_deduplication_id = (
            None
            if not is_fifo
            else kwargs.get("message_deduplication_id", str(uuid.uuid4()))
        )

        if message.manifest:
            message_args["MessageAttributes"] = {
                "manifest": {"DataType": "String", "StringValue": message.manifest},
                "serializer_id": {
                    "DataType": "Number",
                    "StringValue": str(message.serializer_id),
                },
            }

        if kwargs.get("message_group_id"):
            message_args["MessageGroupId"] = kwargs["message_group_id"]

        if message_deduplication_id:
            message_args["MessageDeduplicationId"] = message_deduplication_id

        queue.unwrapped_obj.send_message(**message_args)

    def publish_to_topic(
        self,
        message: Message,
        topic: TopicWrapper,
        extra_attributes: Optional[Dict] = None,
    ) -> None:
        args: Dict = dict(
            Message=message.content,
            MessageAttributes={
                "manifest": {
                    "DataType": "String",
                    "StringValue": message.manifest or "",
                },
                "serializer_id": {
                    "DataType": "Number",
                    "StringValue": str(message.serializer_id),
                },
            },
        )

        if extra_attributes:
            if "subject" in extra_attributes:
                args["Subject"] = extra_attributes["subject"]

            if "message_attributes" in extra_attributes:
                args["MessageAttributes"].update(extra_attributes["message_attributes"])

            if "message_structure" in extra_attributes:
                args["MessageStructure"] = extra_attributes["message_structure"]

        response = topic.unwrapped_obj.publish(**args)

        if "MessageId" not in response:
            raise ConnectionError("Could not send the event to the SNS TOPIC")

    def acknowledge(self, message: Message) -> None:
        message.metadata.delete()

    def close_connection(self) -> None:
        pass

    def delete_queue(self, queue: QueueWrapper) -> None:
        queue.unwrapped_obj.delete()

    def delete_topic(self, topic: TopicWrapper) -> None:
        topic.unwrapped_obj.delete()

    def _construct_message(self, message: Any) -> Message:
        body = message.body
        manifest: Optional[str] = None
        serializer_id: Optional[int] = None
        try:
            message_content = json.loads(body)
            if "Message" in message_content:
                content = message_content["Message"]
                # Does the content have more attributes? If so, it is very likely that the message came from a non-raw
                # SNS redirection
                if "MessageAttributes" in message_content:

                    manifest = (
                        message_content["MessageAttributes"]
                        .get("manifest", {})
                        .get("Value")
                        or ""
                    )
                    serializer_id = (
                        message_content["MessageAttributes"]
                        .get("serializer_id", {})
                        .get("Value")
                    )
            else:
                content = message_content
        except JSONDecodeError:
            content = body

        try:
            manifest = manifest or message.message_attributes.get("manifest", {}).get(
                "StringValue"
            )
            serializer_id = int(
                serializer_id
                or message.message_attributes.get("serializer_id", {}).get(
                    "StringValue"
                )
            )
        except Exception:
            manifest = None
            serializer_id = None

        return Message(message.message_id, content, message, serializer_id, manifest)


class AWSBackend(BaseSQSBackend):
    """
    Backend to use with AWS
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)



import os

from melange.backends.sqs.localsqs import LocalSQSBackend
from melange.examples.payment_service.consumer import PaymentConsumer
from melange.examples.payment_service.publisher import PaymentPublisher
from melange.examples.payment_service.repository import PaymentRepository
from melange.examples.payment_service.service import PaymentService
from melange.examples.shared import serializer_registry
from melange.message_dispatcher import MessageDispatcher
from melange.publishers import QueuePublisher
from melange.serializers.pickle import PickleSerializer

if __name__ == "__main__":
    serializer = PickleSerializer()
    backend = LocalSQSBackend(
        host=os.environ.get("SQSHOST"), port=os.environ.get("SQSPORT")
    )

    dispatcher = MessageDispatcher(
        serializer_registry,
        backend=backend,
    )

    payment_consumer = PaymentConsumer(
        PaymentService(
            PaymentRepository(),
            PaymentPublisher(QueuePublisher(serializer_registry, backend=backend)),
        )
    )

    dispatcher.attach_consumer(payment_consumer)

    print("Consuming...")
    dispatcher.consume_loop("payment-updates")



from typing import Optional

from singleton import Singleton

from melange.backends.interfaces import MessagingBackend


class BackendManager(metaclass=Singleton):
    def __init__(self) -> None:
        self._backend: Optional[MessagingBackend] = None

    def set_default_backend(
        self,
        backend: MessagingBackend,
    ) -> None:
        """
        Sets the default backend
        Args:
            backend:
        """
        if not isinstance(backend, MessagingBackend):
            raise Exception("Invalid backend supplied")

        self._backend = backend

    def get_default_backend(self) -> MessagingBackend:
        """
        Returns the current default backend
        """
        if not self._backend:
            raise Exception(
                "No backend is registered. Please call 'set_default_backend' prior to getting it"
            )

        return self._backend



from typing import Dict, Optional


class SagaState:
    FAILED = "failed"
    COMPLETED = "completed"
    RUNNING = "running"


class Saga:
    def __init__(self, id: str, data: Optional[Dict] = None) -> None:
        self.id = id
        self.data = data
        self.state = SagaState.RUNNING

    def complete(self) -> None:
        self.state = SagaState.COMPLETED

    def fail(self) -> None:
        self.state = SagaState.FAILED



import logging
from typing import Any, Optional

from melange.backends.backend_manager import BackendManager
from melange.backends.interfaces import MessagingBackend
from melange.models import Message
from melange.serializers.registry import SerializerRegistry

logger = logging.getLogger(__name__)


class TopicPublisher:
    def __init__(
        self,
        serializer_registry: SerializerRegistry,
        backend: Optional[MessagingBackend] = None,
    ) -> None:
        self._backend = backend or BackendManager().get_default_backend()
        self._serializer_registry = serializer_registry

    def publish(self, topic_name: str, data: Any, **extra_attributes: Any) -> None:
        """
        Publishes data to a topic

        Args:
            topic_name: The topic to send the message to.
            data: The data to send to this topic. It will be serialized before sending to the
                topic using the serializers.
            **extra_attributes: Any extra attributes. They will be passed to the backend upon publishing.
        """
        topic = self._backend.declare_topic(topic_name)
        serializer = self._serializer_registry.find_serializer_for(data)
        content = serializer.serialize(data)
        manifest = serializer.manifest(data)

        self._backend.publish_to_topic(
            Message.create(content, manifest, serializer.identifier()),
            topic,
            extra_attributes=extra_attributes,
        )


class QueuePublisher:
    def __init__(
        self,
        serializer_registry: SerializerRegistry,
        backend: Optional[MessagingBackend] = None,
    ) -> None:
        self._backend = backend or BackendManager().get_default_backend()
        self._serializer_registry = serializer_registry

    def publish(self, queue_name: str, data: Any, **kwargs: Any) -> None:
        """
        Publishes data to a queue

        Args:
            queue_name: The queue to send the message to.
            data: The data to send to this queue. It will be serialized before sending to the
                queue using the serializers.
            **kwargs: Any extra attributes. They will be passed to the backend upon publish.
        """
        serializer = self._serializer_registry.find_serializer_for(data)
        content = serializer.serialize(data)
        manifest = serializer.manifest(data)
        queue = self._backend.get_queue(queue_name)

        self._backend.publish_to_queue(
            Message.create(content, manifest, serializer.identifier()), queue, **kwargs
        )



from simple_cqrs.domain_event import DomainEvent

from melange.backends.sqs.localsqs import LocalSQSBackend
from melange.publishers import QueuePublisher
from melange.serializers.pickle import PickleSerializer
from melange.serializers.registry import SerializerRegistry


class MyTestMessage:
    def __init__(self, message: str) -> None:
        self.message = message


if __name__ == "__main__":
    serializer_settings = {
        "serializers": {"pickle": PickleSerializer},
        "serializer_bindings": {DomainEvent: "pickle"},
    }

    serializer_registry = SerializerRegistry(serializer_settings)

    backend = LocalSQSBackend(host="localhost", port=9324)
    serializer = PickleSerializer()
    publisher = QueuePublisher(serializer_registry, backend)
    message = MyTestMessage("Hello World!")
    publisher.publish("melangetutorial-queue", message)
    print("Message sent successfully!")



import uuid
from datetime import datetime

from melange.examples.common.commands import DoPayment
from melange.examples.common.domain import Payment
from melange.examples.payment_service.events import OrderResponse, OrderStatus
from melange.examples.payment_service.publisher import PaymentPublisher
from melange.examples.payment_service.repository import PaymentRepository


class PaymentService:
    def __init__(
        self, payment_repository: PaymentRepository, payment_publisher: PaymentPublisher
    ) -> None:
        self.payment_repository = payment_repository
        self.payment_publisher = payment_publisher

    def process(self, order_response: OrderResponse) -> None:
        payment = Payment(
            str(uuid.uuid4()), order_response.id, "SUCCESS", datetime.now()
        )
        self.payment_repository.save(payment)

        order_status = OrderStatus(
            order_id=payment.order_id, payment_id=payment.id, status="ORDER_PAID"
        )
        print(f"Order {order_response.reference} paid successfully")
        self.payment_publisher.publish_orderstatus(order_status)

    def do_payment(self, do_payment: DoPayment) -> None:
        payment = Payment(
            str(uuid.uuid4()), do_payment.order_id, "SUCCESS", datetime.now()
        )
        self.payment_repository.save(payment)

        order_status = OrderStatus(
            order_id=payment.order_id,
            payment_id=payment.id,
            status="ORDER_PAID",
            correlation_id=do_payment.correlation_id,
        )
        print(f"Order {do_payment.reference} paid successfully")
        self.payment_publisher.publish_orderstatus(order_status)



import os

from melange.backends.sqs.localsqs import LocalSQSBackend
from melange.examples.saga_pattern.consumer import SagaConsumer
from melange.examples.saga_pattern.publisher import SagaPublisher
from melange.examples.saga_pattern.repository import SagaRepository
from melange.examples.shared import serializer_registry
from melange.message_dispatcher import SimpleMessageDispatcher
from melange.publishers import QueuePublisher
from melange.serializers.pickle import PickleSerializer

if __name__ == "__main__":
    serializer = PickleSerializer()
    backend = LocalSQSBackend(
        host=os.environ.get("SQSHOST"), port=os.environ.get("SQSPORT")
    )

    payment_consumer = SagaConsumer(
        SagaRepository(),
        SagaPublisher(QueuePublisher(serializer_registry, backend)),
    )

    print("Consuming...")
    consumer_handler = SimpleMessageDispatcher(
        payment_consumer,
        serializer_registry,
        backend=backend,
        always_ack=True,
    )
    consumer_handler.consume_loop("saga-updates")



from abc import ABC, abstractmethod
from typing import Generic, Optional

from melange.helpers.typing import T


class Serializer(ABC, Generic[T]):
    """
    Base interface to inherit for all the serializers
    """

    @staticmethod
    @abstractmethod
    def identifier() -> int:
        raise NotImplementedError

    def manifest(self, data: T) -> Optional[str]:
        """
        Given an object which represents some data, return the
        manifest of that object (if any)
        Args:
            data:

        Returns:
            The manifest of the object
        """
        return None

    def deserialize(self, data: str, manifest: Optional[str] = None) -> T:
        """
        Deserializes a data string. The manifest helps the serializer by
        providing information on how this object must be deserialized.
        Args:
            data: the data string to deserialize
            manifest: the manifest that corresponds to the serialized string

        Returns:
            The data object
        """
        pass

    def serialize(self, data: T) -> str:
        """
        Serializes and object to a string representation
        Args:
            data: the data object to serialize

        Returns:
            A string representation of the `data` object made with this serializer
        """
        pass



from typing import Any, Optional, Protocol

from redis_cache import SimpleCache, logging

logger = logging.getLogger(__name__)


class DeduplicationCache(Protocol):
    def store(self, key: str, value: Any, expire: Optional[int] = None) -> None:
        """
        Stores a key into the cache
        Args:
            key:
            value:
            expire: expiration time in seconds
        """
        raise NotImplementedError

    def get(self, key: str) -> Any:
        """
        Retrieves the value stored under a key
        Args:
            key: the key to fetch in the cache

        Returns:
            The value stored under that key, or None
        """
        raise NotImplementedError

    def __contains__(self, key: str) -> bool:
        """
        Checks whether a certain key is present in the store

        Args:
            key: the key to check

        Returns:
            `True` if the value is present, `False` otherwise.
        """
        raise NotImplementedError


class NullCache:
    """
    A cache that does nothing. Follows the Null Object Pattern.
    """

    def store(self, key: str, value: Any, expire: Optional[int] = None) -> None:
        pass

    def get(self, key: str) -> Any:
        return None

    def __contains__(self, key: str) -> bool:
        return False


class RedisCache:
    def __init__(self, **kwargs: Any) -> None:
        self.cache = SimpleCache(
            expire=kwargs.get("expire", 3600),
            host=kwargs.get("host", 3600),
            port=kwargs.get("port", 3600),
            db=kwargs.get("db", 3600),
            password=kwargs.get("password", 3600),
            namespace=kwargs.get("namespace", 3600),
        )

        if not self.cache.connection:
            logger.warning(
                "Could not establish a connection with redis. Message deduplication won't work"
            )

    def store(self, key: str, value: Any, expire: Optional[int] = None) -> None:
        if not self.cache.connection:
            return

        return self.cache.store(key, value, expire)

    def get(self, key: str) -> Any:
        if not self.cache.connection:
            return

        return self.cache.get(key)

    def __contains__(self, key: str) -> bool:
        return self.cache.connection and key in self.cache



from typing import Any, Dict, Iterable, List, Optional, Tuple

from melange.backends.interfaces import MessagingBackend
from melange.models import Message, QueueWrapper, TopicWrapper


class KafkaBackend(MessagingBackend):
    """
    Kafka works differently than, say, SQS. You connect with the Kafka consumer
    to a list of topics, and then you iterate over them to obtain the messages
    """

    def declare_topic(self, topic_name: str) -> TopicWrapper:
        pass

    def get_queue(self, queue_name: str) -> QueueWrapper:
        pass

    def declare_queue(
        self,
        queue_name: str,
        *topics_to_bind: TopicWrapper,
        dead_letter_queue_name: Optional[str] = None,
        **kwargs: Any
    ) -> Tuple[QueueWrapper, Optional[QueueWrapper]]:
        pass

    def retrieve_messages(self, queue: QueueWrapper, **kwargs: Any) -> List[Message]:
        pass

    def yield_messages(self, queue: QueueWrapper, **kwargs: Any) -> Iterable[Message]:
        pass

    def publish_to_topic(
        self,
        message: Message,
        topic: TopicWrapper,
        extra_attributes: Optional[Dict] = None,
    ) -> None:
        pass

    def publish_to_queue(
        self, message: Message, queue: QueueWrapper, **kwargs: Any
    ) -> None:
        pass

    def acknowledge(self, message: Message) -> None:
        pass

    def close_connection(self) -> None:
        super().close_connection()

    def delete_queue(self, queue: QueueWrapper) -> None:
        pass

    def delete_topic(self, topic: TopicWrapper) -> None:
        pass

    def __init__(self, config: Dict) -> None:
        super().__init__()
        self.config = config



from typing import Dict, Optional

from melange.examples.common.domain import Payment


class PaymentRepository:
    def __init__(self) -> None:
        self.payments: Dict[str, Payment] = {}

    def save(self, payment: Payment) -> None:
        self.payments[payment.id] = payment

    def get(self, id: str) -> Optional[Payment]:
        return self.payments.get(id)



from melange.examples.payment_service.events import OrderStatus
from melange.publishers import QueuePublisher


class PaymentPublisher:
    def __init__(self, publisher: QueuePublisher) -> None:
        self.publisher = publisher

    def publish_orderstatus(self, orderStatus: OrderStatus) -> None:
        self.publisher.publish("order-updates", orderStatus)
        self.publisher.publish("saga-updates", orderStatus)



from melange.backends.factory import MessagingBackendFactory
from melange.backends.sqs.localsqs import LocalSQSBackend

backend = LocalSQSBackend()
factory = MessagingBackendFactory(backend)
factory.init_queue(
    "payment-updates.fifo",
    ["my-topic-1", "my-topic-2", "my-topic-3"],
    dead_letter_queue_name="payment-updates.fifo",
)



from melange.backends.sqs.localsqs import LocalSQSBackend
from melange.examples.shared import serializer_registry
from melange.publishers import TopicPublisher
from melange.serializers.pickle import PickleSerializer


class MyTestMessage:
    def __init__(self, message: str) -> None:
        self.message = message


if __name__ == "__main__":
    backend = LocalSQSBackend(host="localhost", port=9324)
    serializer = PickleSerializer()
    publisher = TopicPublisher(serializer_registry, backend)
    message = MyTestMessage("Hello World!")
    publisher.publish("melangetutorial-topic", message)
    print("Message sent successfully!")



import argparse
import os
import uuid

from melange.backends.sqs.localsqs import LocalSQSBackend
from melange.examples.payment_service.events import OrderResponse
from melange.examples.shared import serializer_registry
from melange.publishers import QueuePublisher
from melange.serializers.pickle import PickleSerializer

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Triggers an order event for the payment service to collect"
    )
    parser.add_argument("reference", type=str, help="The reference number of the order")
    args = parser.parse_args()

    order_reponse = OrderResponse(id=str(uuid.uuid4()), reference=args.reference)
    serializer = PickleSerializer()
    backend = LocalSQSBackend(
        host=os.environ.get("SQSHOST"), port=os.environ.get("SQSPORT")
    )

    QueuePublisher(serializer_registry, backend=backend).publish(
        "saga-updates", order_reponse
    )

    print(f"Order with reference {args.reference} sent.")



import uuid
from typing import Any

from melange.consumers import SingleDispatchConsumer, consumer
from melange.examples.common.commands import DoPayment
from melange.examples.payment_service.events import OrderResponse, OrderStatus
from melange.examples.saga_pattern.models import Saga
from melange.examples.saga_pattern.publisher import SagaPublisher
from melange.examples.saga_pattern.repository import SagaRepository


class SagaConsumer(SingleDispatchConsumer):
    def __init__(
        self,
        saga_repository: SagaRepository,
        saga_publisher: SagaPublisher,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.saga_repository = saga_repository
        self.saga_publisher = saga_publisher

    @consumer
    def on_order_response(self, event: OrderResponse) -> None:
        print(f"Sending payment for order {event.reference}")
        saga = Saga(
            str(uuid.uuid4()),
            data={"order_id": event.id},
        )
        self.saga_repository.save(saga)

        command = DoPayment(
            order_id=event.id, reference=event.reference, correlation_id=saga.id
        )
        self.saga_publisher.publish_dopayment(command)

    @consumer
    def on_order_status(self, event: OrderStatus) -> None:
        print(f"Order received with status: {event.status}")
        saga = self.saga_repository.get(event.correlation_id)
        if not saga:
            return

        print(f"Marking saga {event.correlation_id} as COMPLETED")
        saga.complete()
        self.saga_repository.save(saga)



from simple_cqrs.domain_event import DomainEvent

from melange.serializers.pickle import PickleSerializer
from melange.serializers.registry import SerializerRegistry

serializer_settings = {
    "serializers": {"pickle": PickleSerializer},
    "serializer_bindings": {DomainEvent: "pickle"},
}

serializer_registry = SerializerRegistry(serializer_settings)



import json
from typing import Dict, Optional

from melange.serializers.interfaces import Serializer


class JsonSerializer(Serializer[Dict]):
    """
    Serializes and deserializes python dictionaries in json format
    """

    @staticmethod
    def identifier() -> int:
        return 1

    def manifest(self, data: Dict) -> str:
        return "json"

    def deserialize(self, serialized_data: str, manifest: Optional[str] = None) -> Dict:
        data = json.loads(serialized_data)
        return data

    def serialize(self, data: Dict) -> str:
        return json.dumps(data)



from melange.examples.common.commands import DoPayment
from melange.publishers import QueuePublisher


class SagaPublisher:
    def __init__(self, publisher: QueuePublisher) -> None:
        self.publisher = publisher

    def publish_dopayment(self, do_payment: DoPayment) -> None:
        self.publisher.publish("payment-updates", do_payment)



from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

from melange.backends.interfaces import MessagingBackend
from melange.consumers import Consumer
from melange.message_dispatcher import MessageDispatcher
from melange.models import Message, QueueWrapper, TopicWrapper
from melange.serializers.registry import SerializerRegistry


def link_synchronously(
    queue_name: str,
    consumers: List[Consumer],
    serializer_registry: SerializerRegistry,
    backend: "InMemoryMessagingBackend",
) -> None:
    consumer_dispatcher = MessageDispatcher(serializer_registry, backend=backend)
    for consumer in consumers:
        consumer_dispatcher.attach_consumer(consumer)

    backend.set_callback(queue_name, consumer_dispatcher.consume_event)


class DumbObject:
    def __init__(self, name: str) -> None:
        self.name = name


class InMemoryMessagingBackend(MessagingBackend):
    def __init__(self) -> None:
        super().__init__()
        self._messages: List[Message] = []
        self.callbacks: Dict = {}

    def set_callback(self, queue_name: str, callback: Callable[[str], None]) -> None:
        self.callbacks[queue_name] = callback

    def declare_topic(self, topic_name: str) -> TopicWrapper:
        return TopicWrapper(DumbObject(topic_name))

    def get_queue(self, queue_name: str) -> QueueWrapper:
        return QueueWrapper(DumbObject(queue_name))

    def declare_queue(
        self,
        queue_name: str,
        *topics_to_bind: TopicWrapper,
        dead_letter_queue_name: Optional[str] = None,
        **kwargs: Any
    ) -> Tuple[QueueWrapper, Optional[QueueWrapper]]:
        return QueueWrapper(DumbObject(queue_name)), None

    def retrieve_messages(self, queue: QueueWrapper, **kwargs: Any) -> List[Message]:
        messages = self._messages
        self._messages = []
        return messages

    def yield_messages(self, queue: QueueWrapper, **kwargs: Any) -> Iterable[Message]:
        for message in self._messages:
            yield message

    def publish_to_topic(
        self,
        message: Message,
        topic: TopicWrapper,
        extra_attributes: Optional[Dict] = None,
    ) -> None:
        if topic.unwrapped_obj.name in self.callbacks:
            # Append the message internally to simulate "the queue",
            # then call the
            self._messages.append(message)

            self.callbacks[topic.unwrapped_obj.name](topic.unwrapped_obj.name)

    def publish_to_queue(
        self, message: Message, queue: QueueWrapper, **kwargs: Any
    ) -> None:
        if queue.unwrapped_obj.name in self.callbacks:
            # Append the message internally to simulate "the queue",
            # then call the
            self._messages.append(message)

            self.callbacks[queue.unwrapped_obj.name](queue.unwrapped_obj.name)

    def acknowledge(self, message: Message) -> None:
        return None

    def delete_queue(self, queue: QueueWrapper) -> None:
        return None

    def delete_topic(self, topic: TopicWrapper) -> None:
        return None



import logging
from typing import Any, Callable, Optional

from methoddispatch import SingleDispatch, singledispatch

logger = logging.getLogger(__name__)


class Consumer:
    """
    A consumer is responsible to process a message and do
    whatever is needed by the application.

    You could use is as is and supply a `on_message` callable to process your messages.
    Though commonly you would inherit this class, create your own consumer and
    override the `process` and `accepts` methods.
    """

    def __init__(self, on_message: Optional[Callable[[Any], None]] = None) -> None:
        self._on_message = on_message

    def process(self, message: Any, **kwargs: Any) -> None:
        """
        Processes a message
        Args:
            message: the message data to process, already deserialied
            **kwargs: any other parameters the dispatcher sends upon processing
        """
        if self._on_message:
            self._on_message(message)

    def accepts(self, manifest: Any) -> bool:
        """
        Determines whether it is able to handle a message or not
        """
        return True


class SingleDispatchConsumer(Consumer, SingleDispatch):
    """
    This class can consume events from a queue and pass them to a processor
    through the means of method overloading. Provides a default implementation as well for
    the accepts method
    """

    def process(self, message: Any, **kwargs: Any) -> None:
        self._process(message)

    @singledispatch
    def _process(self, message: Any) -> None:
        """Event should be an instance of DomainEvent"""
        pass

    def accepts(self, message: Any) -> bool:
        accepted_types = filter(lambda t: t is not object, self._process.registry)
        return any(isinstance(message, t) for t in accepted_types)


consumer = SingleDispatchConsumer._process.register



from typing import Any, Optional


class TopicWrapper:
    """
    Wrapper class for a topic of any infrastructure technology.
    It's intended to only be unwrapped at the messaging backend
    who is the only one responsible to handle the specifics
    """

    def __init__(self, obj: Any) -> None:
        self.obj = obj

    @property
    def unwrapped_obj(self) -> Any:
        return self.obj

    def unwrap(self) -> Any:
        return self.obj


class QueueWrapper:
    """
    Wrapper class for a queue of any infrastructure technology
    """

    def __init__(self, obj: Any) -> None:
        self.obj = obj

    @property
    def unwrapped_obj(self) -> Any:
        return self.obj

    def unwrap(self) -> Any:
        return self.obj


class Message:
    def __init__(
        self,
        message_id: Optional[str],
        content: str,
        metadata: Any,
        serializer_id: Optional[int],
        manifest: Optional[str] = None,
    ) -> None:
        self.message_id = message_id
        self.content = content
        self.metadata = metadata
        self.serializer_id = serializer_id
        self.manifest = manifest

    @staticmethod
    def create(
        content: str, manifest: Optional[str], serializer_id: int, metadata: Any = None
    ) -> "Message":
        return Message(None, content, metadata, serializer_id, manifest)

    def get_message_manifest(self) -> Optional[str]:
        return self.manifest



from base64 import b64decode
from typing import Dict, List
from urllib.request import urlopen

import OpenSSL
from OpenSSL.crypto import FILETYPE_PEM, load_certificate, verify

SNS_MESSAGE_TYPE_SUB_NOTIFICATION = "SubscriptionConfirmation"
SNS_MESSAGE_TYPE_NOTIFICATION = "Notification"
SNS_MESSAGE_TYPE_UNSUB_NOTIFICATION = "UnsubscribeConfirmation"


def canonical_message_builder(content: Dict, fmt: List[str]) -> str:
    """Builds the canonical message to be verified.

    Sorts the fields as a requirement from AWS

    Args:
        content (dict): Parsed body of the response
        fmt (list): List of the fields that need to go into the message
    Returns (str):
        canonical message
    """
    m = ""

    for field in sorted(fmt):
        try:
            m += field + "\n" + content[field] + "\n"
        except KeyError:
            # Build with what you have
            pass

    return str(m)


def verify_sns_notification(message: Dict) -> bool:
    """Takes a notification request from Amazon push service SNS and verifies the origin of the notification.
    This implementation uses OpenSSL Crypto, inspired by the implementation of
    Artur Rodrigues with M2Crypto: http://goo.gl/KAgPPc
    Args:
        message (dictionary):
            The parsed message content
    Returns (bool):
        True if he message passes the verification, False otherwise
    """

    canonical_sub_unsub_format = [
        "Message",
        "MessageId",
        "SubscribeURL",
        "Timestamp",
        "Token",
        "TopicArn",
        "Type",
    ]
    canonical_notification_format = [
        "Message",
        "MessageId",
        "Subject",
        "Timestamp",
        "TopicArn",
        "Type",
    ]

    content = message
    decoded_signature = b64decode(content["Signature"])

    # Depending on the message type, canonical message format varies: http://goo.gl/oSrJl8
    if (
        message["Type"] == SNS_MESSAGE_TYPE_SUB_NOTIFICATION
        or message["Type"] == SNS_MESSAGE_TYPE_UNSUB_NOTIFICATION
    ):

        canonical_message = canonical_message_builder(
            content, canonical_sub_unsub_format
        )

    elif message["Type"] == SNS_MESSAGE_TYPE_NOTIFICATION:

        canonical_message = canonical_message_builder(
            content, canonical_notification_format
        )

    else:
        raise ValueError("Message Type (%s) is not recognized" % message["Type"])

    # Load the certificate and extract the public key
    cert = load_certificate(
        FILETYPE_PEM, urlopen(content["SigningCertURL"]).read().decode("utf-8")
    )

    try:
        verify(cert, decoded_signature, str.encode(canonical_message), "sha1")
    except OpenSSL.crypto.Error:
        return False

    return True



import os

from melange.backends.factory import MessagingBackendFactory
from melange.backends.sqs.localsqs import LocalSQSBackend

if __name__ == "__main__":
    backend = LocalSQSBackend(
        host=os.environ.get("SQSHOST"), port=os.environ.get("SQSPORT")
    )

    factory = MessagingBackendFactory(
        backend=backend,
    )
    factory.init_queue("payment-updates")
    factory.init_queue("order-updates")
    factory.init_queue("saga-updates")

    print("Queues created.")



from melange.consumers import SingleDispatchConsumer, consumer
from melange.examples.common.commands import DoPayment
from melange.examples.payment_service.events import OrderResponse
from melange.examples.payment_service.service import PaymentService


class PaymentConsumer(SingleDispatchConsumer):
    def __init__(self, payment_service: PaymentService):
        super().__init__()
        self.payment_service = payment_service

    @consumer
    def consume_order_response(self, event: OrderResponse) -> None:
        self.payment_service.process(event)

    @consumer
    def consume_do_payment(self, command: DoPayment) -> None:
        self.payment_service.do_payment(command)



from typing import Any

from melange.backends.sqs.sqs_backend import BaseSQSBackend


class LocalSQSBackend(BaseSQSBackend):
    """
    Local backend to use with elasticMQ
    """

    def __init__(self, **kwargs: Any) -> None:

        super().__init__(
            extra_settings=dict(
                endpoint_url=f"{kwargs.get('host', 'localhost')}:{kwargs.get('port', 4566)}",
                region_name="us-east-1",
                aws_secret_access_key="x",
                aws_access_key_id="x",
                use_ssl=False,
            ),
            sns_settings=dict(
                endpoint_url=f"{kwargs.get('sns_host', 'localhost')}:{kwargs.get('sns_port', 4566)}",
                aws_secret_access_key="x",
                aws_access_key_id="x",
                region_name="us-east-1",
                use_ssl=False,
            ),
            **kwargs,
        )



from typing import Any, Optional

from melange.consumers import Consumer
from melange.examples.common.commands import DoPayment
from melange.examples.payment_service.events import OrderResponse
from melange.examples.payment_service.service import PaymentService


class PaymentConsumer(Consumer):
    def __init__(self, payment_service: PaymentService):
        super().__init__()
        self.payment_service = payment_service

    def process(self, message: Any, **kwargs: Any) -> None:
        if isinstance(message, OrderResponse):
            self.payment_service.process(message)
        elif isinstance(message, DoPayment):
            self.payment_service.do_payment(message)

    def accepts(self, manifest: Optional[str]) -> bool:
        return manifest in ["OrderResponse", "DoPayment"]



import logging
from typing import Any, Callable, List, Optional

from melange.backends.backend_manager import BackendManager
from melange.backends.interfaces import MessagingBackend
from melange.consumers import Consumer
from melange.exceptions import SerializationError
from melange.infrastructure.cache import DeduplicationCache, NullCache
from melange.models import Message
from melange.serializers.registry import SerializerRegistry
from melange.utils import get_fully_qualified_name

# TODO: Clear code duplication between the dispatcher and the handler. Fix.

logger = logging.getLogger(__name__)


class MessageDispatcher:
    """
    The `MessageDispatcher` is responsible to start the message consumption loop
    to retrieve the available messages from the queue and dispatch them to the
    consumers.
    """

    def __init__(
        self,
        serializer_registry: SerializerRegistry,
        cache: Optional[DeduplicationCache] = None,
        backend: Optional[MessagingBackend] = None,
        always_ack: bool = False,
    ) -> None:
        self._consumers: List[Consumer] = []
        self.serializer_registry = serializer_registry
        self._backend = backend or BackendManager().get_default_backend()
        self.cache: DeduplicationCache = cache or NullCache()
        self.always_ack = always_ack

    def attach_consumer(self, consumer: Consumer) -> None:
        """
        Attaches a consumer to the dispatcher, so that it can receive messages
        Args:
            consumer: the consumer to attach
        """
        if consumer not in self._consumers:
            self._consumers.append(consumer)

    def unattach_consumer(self, consumer: Consumer) -> None:
        """
        Unattaches the consumer from the dispatcher, so that it will not receive messages anymore
        Args:
            consumer: the consumer to unattach
        """
        if consumer in self._consumers:
            self._consumers.remove(consumer)

    def consume_loop(
        self,
        queue_name: str,
        on_exception: Optional[Callable[[Exception], None]] = None,
        after_consume: Optional[Callable[[], None]] = None,
    ) -> None:
        """
        Starts the consumption loop on the queue `queue_name`
        Args:
            queue_name: The queue to poll for new messages
            on_exception: If there is any exception, the exception
                will be passed to this callback
            after_consume: After consuming a batch of events,
                invoke this callback
        """
        self.consume_event(queue_name, on_exception, after_consume)

    def consume_event(
        self,
        queue_name: str,
        on_exception: Optional[Callable[[Exception], None]] = None,
        after_consume: Optional[Callable[[], None]] = None,
    ) -> None:
        """
        Consumes events on the queue `queue_name`
        """
        event_queue = self._backend.get_queue(queue_name)

        messages = self._backend.yield_messages(event_queue)
        for message in messages:
            try:
                self._dispatch_message(message)
            except Exception as e:
                logger.exception(e)
                if on_exception:
                    on_exception(e)
            finally:
                if after_consume:
                    after_consume()

    def _get_consumers(self, message_data: Any) -> List[Consumer]:
        return [
            consumer for consumer in self._consumers if consumer.accepts(message_data)
        ]

    def _dispatch_message(self, message: Message) -> None:
        manifest = message.get_message_manifest()

        # If the message cannot be deserialized, just ignore it.
        # ACK it anyway to avoid hanging on the same message over an over again
        try:
            if message.serializer_id is not None:
                message_data = self.serializer_registry.deserialize_with_serializerid(
                    message.content, message.serializer_id, manifest=manifest
                )
            else:
                # TODO: If no serializerid is supplied, at least we need to grab some
                # clue in the message to know which serializer must be used.
                # for now, rely on the default.
                message_data = self.serializer_registry.deserialize_with_class(
                    message.content, object, manifest=manifest
                )
        except SerializationError as e:
            logger.error(e)
            self._backend.acknowledge(message)
            return

        consumers = self._get_consumers(message_data)

        successful = 0
        for consumer in consumers:
            try:
                # Store into the cache
                message_key = (
                    f"{get_fully_qualified_name(consumer)}.{message.message_id}"
                )

                if message_key in self.cache:
                    logger.info("detected a duplicated message, ignoring")
                else:
                    consumer.process(message_data, message_id=message.message_id)
                    successful += 1
                    self.cache.store(message_key, message_key)
            except Exception as e:
                logger.exception(e)

        if self.always_ack or successful == len(consumers):
            self._backend.acknowledge(message)


class SimpleMessageDispatcher(MessageDispatcher):
    """
    Handles the retrieval of messages from a queue and
    dispatches the messages to a single consumer. A simpler version
    of the `MessageDispatcher`
    """

    def __init__(
        self,
        consumer: Consumer,
        serializer_registry: SerializerRegistry,
        cache: Optional[DeduplicationCache] = None,
        backend: Optional[MessagingBackend] = None,
        always_ack: bool = False,
    ):
        super().__init__(serializer_registry, cache, backend, always_ack)
        self.attach_consumer(consumer)



from melange.backends.factory import MessagingBackendFactory
from melange.backends.sqs.localsqs import LocalSQSBackend

backend = LocalSQSBackend()
factory = MessagingBackendFactory(backend)
factory.init_queue("payment-updates.fifo")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['melange',
 'melange.backends',
 'melange.backends.rabbitmq',
 'melange.backends.sqs',
 'melange.examples',
 'melange.examples.common',
 'melange.examples.doc_examples',
 'melange.examples.doc_examples.tutorial',
 'melange.examples.payment_service',
 'melange.examples.saga_pattern',
 'melange.helpers',
 'melange.infrastructure',
 'melange.serializers']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.11.5,<2.0.0',
 'funcy>=1.14,<2.0',
 'marshmallow>=3.3.0,<4.0.0',
 'methoddispatch>=3.0.2,<4.0.0',
 'pika>=1.1.0,<2.0.0',
 'pytz>=2019.3,<2020.0',
 'redis-simple-cache-py3>=0.0.7,<0.0.8',
 'singleton-py3>=0.2.1,<0.3.0',
 'toolz>=0.11.2,<0.12.0']

setup_kwargs = {
    'name': 'melange',
    'version': '7.1.0',
    'description': 'A messaging library for an easy inter-communication in distributed and microservices architectures',
    'long_description': None,
    'author': 'David Jiménez (Rydra)',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<4.0.0',
}


setup(**setup_kwargs)



import json
from typing import Dict, Optional

from bunch import Bunch

from melange.exceptions import SerializationError
from melange.helpers.typing import T
from melange.serializers.interfaces import Serializer


class MyOwnSerializer(Serializer):
    # If you define a constructor for you serializer, it should accept no parameters

    @staticmethod
    def identifier() -> int:
        """
        Pick a unique identifier for your Serializer.
        The numbers 0 to 40 are reserved for Melange itself
        """
        return 123

    def manifest(self, data: T) -> Optional[str]:
        """
        The manifest is a type hint used on deserialization
        """
        if isinstance(data, Bunch):
            return "bunch"
        if isinstance(data, Dict):
            return "dictionary"
        else:
            raise SerializationError(
                "The data object cannot be used with this serializer"
            )

    def deserialize(self, data: str, manifest: Optional[str] = None) -> T:
        """
        `deserialize` uses the manifest as a type hint
        """
        if manifest == "dictionary":
            deserialized_data = json.loads(data)
            return deserialized_data
        elif manifest == "bunch":
            deserialized_data = Bunch(json.loads(data))
            return deserialized_data
        else:
            raise SerializationError("Unknown manifest")

    def serialize(self, data: T) -> str:
        """
        Serializes the given object to an string
        """
        if isinstance(data, Bunch):
            return json.dumps(dict(data))
        elif isinstance(data, Dict):
            return json.dumps(data)
        else:
            raise SerializationError("This type cannot be used with this serializer")



from melange.backends.factory import MessagingBackendFactory
from melange.backends.sqs.localsqs import LocalSQSBackend

backend = LocalSQSBackend(host="localhost", port=9324)
factory = MessagingBackendFactory(backend)
factory.init_queue("melangetutorial-queue")
